//		Copyright © 1998–1999  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//		$Log: SCSIData.h,v $
//		Revision 1.6  1999/04/18 13:19:19  jjk
//		Get rid of virtual functions (Intel support).  Arrgh!
//
//		Revision 1.5  1999/04/18 10:13:21  jjk
//		GCC support; add LOCATE command.
//
//		Revision 1.4  1998/10/23 09:13:11  jjk
//		Intel support (untested).
//
//		Revision 1.3  1998/06/11 22:12:36  jjk
//		Fix a wrong conversion.
//
//		Revision 1.2  1998/06/07 16:25:50  jjk
//		Add utilities to convert between integers and byte arrays.
//
//		Revision 1.1  1998/06/06 15:16:35  jjk
//		Initial revision
//

//		Data structures for SCSI commands and data.

#ifndef ST_SCSIDATA_H
#define ST_SCSIDATA_H

#include "SCSIDefs.h"

//------------------------------------------------------------------------------
// Macros for specifying bitfields in a portable way (using an idea from
// the Mach SCSI driver).

#if defined(__POWERPC__)
#  define BF_MSB_FIRST 1
#elif defined(__INTEL__)
#  define BF_MSB_FIRST 0
#else
#  error Must define BF_MSB_FIRST to 0 or 1.
#endif

#if BF_MSB_FIRST
#  define B2(a, b)							a, b
#  define B3(a, b, c)						a, b, c
#  define B4(a, b, c, d)					a, b, c, d
#  define B5(a, b, c, d, e)				a, b, c, d, e
#  define B6(a, b, c, d, e, f)			a, b, c, d, e, f
#  define B7(a, b, c, d, e, f, g)		a, b, c, d, e, f, g
#  define B8(a, b, c, d, e, f, g, h)	a, b, c, d, e, f, g, h
#else
#  define B2(a, b)							b, a
#  define B3(a, b, c)						c, b, a
#  define B4(a, b, c, d)					d, c, b, a
#  define B5(a, b, c, d, e)				e, d, c, b, a
#  define B6(a, b, c, d, e, f)			f, e, d, c, b, a
#  define B7(a, b, c, d, e, f, g)		g, f, e, d, c, b, a
#  define B8(a, b, c, d, e, f, g, h)	h, g, f, e, d, c, b, a
#endif

#define BF2(t, a, b)							t B2(a, b)
#define BF3(t, a, b, c)						t B3(a, b, c)
#define BF4(t, a, b, c, d)					t B4(a, b, c, d)
#define BF5(t, a, b, c, d, e)				t B5(a, b, c, d, e)
#define BF6(t, a, b, c, d, e, f)			t B6(a, b, c, d, e, f)
#define BF7(t, a, b, c, d, e, f, g)		t B7(a, b, c, d, e, f, g)
#define BF8(t, a, b, c, d, e, f, g, h)	t B8(a, b, c, d, e, f, g, h)

//------------------------------------------------------------------------------
// Base class for SCSI data representation.
class SCSIData
{
public:
	// Methods to convert between integers and 2/3/4-byte arrays.
	static uint16 u2b_i16(const uint8 data[2])
	{
		return ((uint16)data[0] << 8)
				|((uint16)data[1]);
	};

	static int16 s2b_i16(const uint8 data[2])
	{
		return (int16)u2b_i16(data);
	};

	static uint32 u3b_i32(const uint8 data[3])
	{
		return ((uint32)data[0] << 16)
				|((uint32)data[1] << 8)
				|((uint32)data[2]);
	};

	static int32 s3b_i32(const uint8 data[3])
	{
		return (int32)u3b_i32(data);
	};

	static uint32 u4b_i32(const uint8 data[4])
	{
		return ((uint32)data[0] << 24)
				|((uint32)data[1] << 16)
				|((uint32)data[2] << 8)
				|((uint32)data[3]);
	};

	static int32 s4b_i32(const uint8 data[4])
	{
		return (int32)u4b_i32(data);
	};
	
	static void ui16_2b(const uint16 value, uint8 data[2])
	{
		data[0] = value >> 8;
		data[1] = value;
	};

	static void si16_2b(const int16 value, uint8 data[2])
	{
		ui16_2b((uint16)value, data);
	};
	
	static void ui32_3b(const uint32 value, uint8 data[3])
	{
		data[0] = value >> 16;
		data[1] = value >> 8;
		data[2] = value;
	};

	static void si32_3b(const int32 value, uint8 data[3])
	{
		ui32_3b((uint32)value, data);
	};
	
	static void ui32_4b(const uint32 value, uint8 data[4])
	{
		data[0] = value >> 24;
		data[1] = value >> 16;
		data[2] = value >> 8;
		data[3] = value;
	};

	static void si32_4b(const int32 value, uint8 data[4])
	{
		ui32_4b((uint32)value, data);
	};
};

//------------------------------------------------------------------------------
// Basic types for representing SCSI Command Descriptor Blocks.
class CDB : public SCSIData
{
public:
	// Inquire CDB address, opcode and length.
	uint8 *Address(void)
	{
		return mpData;
	}

	uint8 Opcode(void) const
	{
		return mpData[0];
	}
	
	uint8 Length(void) const
	{
		return mLength;
	}
	
protected:
	// Constructor.
	CDB(uint8 *pData, const uint8 length)
	:	mpData(pData),
		mLength(length)
	{
		// empty
	}
	
private:
	// Data members.
	uint8 *const mpData;
	const uint8 mLength;
};

class CDB6 : public CDB
{
protected:
	// Constructor.
	CDB6(uint8 *pData)
	:	CDB(pData, 6)
	{
		// empty
	}
};

class CDB10 : public CDB
{
protected:
	// Constructor.
	CDB10(uint8 *pData)
	:	CDB(pData, 10)
	{
		// empty
	}
};

class CDB12 : public CDB
{
protected:
	// Constructor.
	CDB12(uint8 *pData)
	:	CDB(pData, 12)
	{
		// empty
	}
};

//------------------------------------------------------------------------------
// CDBs for specific SCSI commands.

// ERASE
class CDBErase : public CDB6
{
	typedef CDB6 inherited;
public:
		 uint8 opCode;
	BF4(uint8,lun:3,
				 res1:3,
				 immed:1,
				 lang:1);	// %$#%$# reservierte Wörter!
		 uint8 res2;
		 uint8 res3;
		 uint8 res4;
		 uint8 control;
	
	CDBErase(uint8 lun_,
				bool long_ = true, bool immediate = false,
				uint8 control_ = 0)
	:	inherited(&opCode),
		opCode(SCSI_CMD_ERASE),
		B4(lun(lun_),
			res1(0),
			immed(immediate),
			lang(long_)),
		res2(0),
		res3(0),
		res4(0),
		control(control_)
	{
		// empty
	}
};

// LOAD UNLOAD
class CDBLoadUnload : public CDB6
{
	typedef CDB6 inherited;
public:
		 uint8 opCode;
	BF3(uint8,lun:3,
				 res1:4,
				 immed:1);
		 uint8 res2;
		 uint8 res3;
	BF4(uint8,res4:5,
				 eot:1,
				 reten:1,
				 load:1);
		 uint8 control;

	CDBLoadUnload(uint8 lun_,
					  bool load_, bool flag = false, bool immediate = false,
					  uint8 control_ = 0)
	:	inherited(&opCode),
		opCode(SCSI_CMD_LOAD_UNLOAD),
		B3(lun(lun_),
			res1(0),
			immed(immediate)),
		res2(0),
		res3(0),
		B4(res4(0),
			eot(load_ ? false : flag),
			reten(load_ ? flag : false),
			load(load_)),
		control(control_)
	{
		// empty
	}
};

// LOCATE
class CDBLocate : public CDB10
{
	typedef CDB10 inherited;
public:
	enum {		// Block address types.
		kSCSILogicalBlockAddress = 0,
		kDeviceSpecific			 = 1
	};

		 uint8 opCode;
	BF5(uint8,lun:3,
				 res1:2,
				 bt:1,
				 cp:1,
				 immed:1);
		 uint8 res2;
		 uint8 blockAddress[4];
		 uint8 res7;
		 uint8 partition;
		 uint8 control;

	CDBLocate(uint8 lun_,
				 uint32 blockAddress_,
				 uint8 blockAddressType = kSCSILogicalBlockAddress,
				 uint8 partition_ = 0,
				 bool changePartition = false, bool immediate = false,
				 uint8 control_ = 0)
	:	inherited(&opCode),
		opCode(SCSI_CMD_LOCATE),
		B5(lun(lun_),
			res1(0),
			bt(blockAddressType),
			cp(changePartition),
			immed(immediate)),
		res2(0),
		res7(0),
		partition(partition_),
		control(control_)
	{
		ui32_4b(blockAddress_, blockAddress);
	}
};

// MODE SELECT (6)
class CDBModeSelect0 : public CDB6
{
	typedef CDB6 inherited;
public:
		 uint8 opCode;
	BF4(uint8,lun:3,
				 pf:1,
				 res1:3,
				 sp:1);
		 uint8 res2;
		 uint8 res3;
		 uint8 length;
		 uint8 control;

	CDBModeSelect0(uint8 lun_, uint8 length_,
						bool pageFormat = true, bool savePages = false,
						uint8 control_ = 0)
	:	inherited(&opCode),
		opCode(SCSI_CMD_MODE_SELECT_0),
		B4(lun(lun_),
			pf(pageFormat),
			res1(0),
			sp(savePages)),
		res2(0),
		res3(0),
		length(length_),
		control(control_)
	{
		// empty
	}
};

// MODE SENSE (6)
class CDBModeSense0 : public CDB6
{
	typedef CDB6 inherited;
public:
	enum {		// Page Control field values.
		kCurrentValues		= 0,
		kChangeableValues	= 1,
		kDefaultValues		= 2,
		kSavedValues		= 3
	};
	
		 uint8 opCode;
	BF4(uint8,lun:3,
				 res10:1,
				 dbd:1,
				 res11:3);
	BF2(uint8,pc:2,
				 pageCode:6);
		 uint8 res3;
		 uint8 length;
		 uint8 control;

	CDBModeSense0(uint8 lun_, uint8 pageCode_, uint8 length_,
					  bool disableBlockDescriptors = false,
					  uint8 pageControl = kCurrentValues,
					  uint8 control_ = 0)
	:	inherited(&opCode),
		opCode(SCSI_CMD_MODE_SENSE_0),
		B4(lun(lun_),
			res10(0),
			dbd(disableBlockDescriptors),
			res11(0)),
		B2(pc(pageControl),
			pageCode(pageCode_)),
		res3(0),
		length(length_),
		control(control_)
	{
		// empty
	}
};

// PREVENT/ALLOW MEDIUM REMOVAL
class CDBMediumRemoval : public CDB6
{
	typedef CDB6 inherited;
public:
		 uint8 opCode;
	BF2(uint8,lun:3,
				 res1:5);
		 uint8 res2;
		 uint8 res3;
	BF2(uint8,res4:7,
				 prvnt:1);
		 uint8 control;
	
	CDBMediumRemoval(uint8 lun_, bool prvnt_, uint8 control_ = 0)
	:	inherited(&opCode),
		opCode(SCSI_CMD_PREVENT_ALLOW_REMOVAL),
		B2(lun(lun_),
			res1(0)),
		res2(0),
		res3(0),
		B2(res4(0),
			prvnt(prvnt_)),
		control(control_)
	{
		// empty
	}
};

// READ
class CDBRead : public CDB6
{
	typedef CDB6 inherited;
public:
		 uint8 opCode;
	BF4(uint8,lun:3,
				 res1:3,
				 sili:1,
				 fixed:1);
		 uint8 length[3];
		 uint8 control;
	
	CDBRead(uint8 lun_, uint32 length_,
			  bool fixed_ = false, bool suppressILI = false,
			  uint8 control_ = 0)
	:	inherited(&opCode),
		opCode(SCSI_CMD_READ),
		B4(lun(lun_),
			res1(0),
			sili(suppressILI),
			fixed(fixed_)),
		control(control_)
	{
		ui32_3b(length_, length);
	}
};

// READ BLOCK LIMITS
class CDBReadBlockLimits : public CDB6
{
	typedef CDB6 inherited;
public:
		 uint8 opCode;
	BF2(uint8,lun:3,
				 res1:5);
		 uint8 res2;
		 uint8 res3;
		 uint8 res4;
		 uint8 control;
	
	CDBReadBlockLimits(uint8 lun_, uint8 control_ = 0)
	:	inherited(&opCode),
		opCode(SCSI_CMD_READ_BLOCK_LIMITS),
		B2(lun(lun_),
			res1(0)),
		res2(0),
		res3(0),
		res4(0),
		control(control_)
	{
		// empty
	}
};

// READ POSITION
class CDBReadPosition : public CDB10
{
	typedef CDB10 inherited;
public:
	enum {		// Block address types.
		kSCSILogicalBlockAddress = 0,
		kDeviceSpecific			 = 1
	};

		 uint8 opCode;
	BF3(uint8,lun:3,
				 res1:4,
				 bt:1);
		 uint8 res2;
		 uint8 res3;
		 uint8 res4;
		 uint8 res5;
		 uint8 res6;
		 uint8 res7;
		 uint8 res8;
		 uint8 control;
	
	CDBReadPosition(uint8 lun_,
						 uint8 blockAddressType = kSCSILogicalBlockAddress,
						 uint8 control_ = 0)
	:	inherited(&opCode),
		opCode(SCSI_CMD_READ_POSITION),
		B3(lun(lun_),
			res1(0),
			bt(blockAddressType)),
		res2(0),
		res3(0),
		res4(0),
		res5(0),
		res6(0),
		res7(0),
		res8(0),
		control(control_)
	{
		// empty
	}
};

// REWIND
class CDBRewind : public CDB6
{
	typedef CDB6 inherited;
public:
		 uint8 opCode;
	BF3(uint8,lun:3,
				 res1:4,
				 immed:1);
		 uint8 res2;
		 uint8 res3;
		 uint8 res4;
		 uint8 control;
	
	CDBRewind(uint8 lun_, bool immediate = false, uint8 control_ = 0)
	:	inherited(&opCode),
		opCode(SCSI_CMD_REWIND),
		B3(lun(lun_),
			res1(0),
			immed(immediate)),
		res2(0),
		res3(0),
		res4(0),
		control(control_)
	{
		// empty
	}
};

// SPACE
class CDBSpace : public CDB6
{
	typedef CDB6 inherited;
public:
	enum {		// Code field values.
		kBlocks					= 0,
		kFilemarks				= 1,
		kSequentialFilemarks	= 2,
		kEndOfData				= 3,
		kSetmarks				= 4,
		kSequentialSetmarks	= 5
	};
	
		 uint8 opCode;
	BF3(uint8,lun:3,
				 res1:2,
				 code:3);
		 uint8 count[3];
		 uint8 control;
	
	CDBSpace(uint8 lun_, uint8 code_, int32 count_, uint8 control_ = 0)
	:	inherited(&opCode),
		opCode(SCSI_CMD_SPACE),
		B3(lun(lun_),
			res1(0),
			code(code_)),
		control(control_)
	{
		si32_3b(count_, count);
	}
};

// TEST UNIT READY
class CDBTestUnitReady : public CDB6
{
	typedef CDB6 inherited;
public:
		 uint8 opCode;
	BF2(uint8,lun:3,
				 res1:5);
		 uint8 res2;
		 uint8 res3;
		 uint8 res4;
		 uint8 control;
	
	CDBTestUnitReady(uint8 lun_, uint8 control_ = 0)
	:	inherited(&opCode),
		opCode(SCSI_CMD_TEST_UNIT_READY),
		B2(lun(lun_),
			res1(0)),
		res2(0),
		res3(0),
		res4(0),
		control(control_)
	{
		// empty
	}
};

// VERIFY
class CDBVerify0 : public CDB6
{
	typedef CDB6 inherited;
public:
		 uint8 opCode;
	BF5(uint8,lun:3,
				 res1:2,
				 immed:1,
				 bytcmp:1,
				 fixed:1);
		 uint8 length[3];
		 uint8 control;
	
	CDBVerify0(uint8 lun_, uint32 length_, bool immediate = false,
				  bool byteCompare = false, bool fixed_ = false,
				  uint8 control_ = 0)
	:	inherited(&opCode),
		opCode(SCSI_CMD_VERIFY_0),
		B5(lun(lun_),
			res1(0),
			immed(immediate),
			bytcmp(byteCompare),
			fixed(fixed_)),
		control(control_)
	{
		ui32_3b(length_, length);
	}
};

// WRITE
class CDBWrite : public CDB6
{
	typedef CDB6 inherited;
public:
		 uint8 opCode;
	BF3(uint8,lun:3,
				 res1:4,
				 fixed:1);
		 uint8 length[3];
		 uint8 control;
	
	CDBWrite(uint8 lun_, uint32 length_, bool fixed_ = false,
			  uint8 control_ = 0)
	:	inherited(&opCode),
		opCode(SCSI_CMD_WRITE),
		B3(lun(lun_),
			res1(0),
			fixed(fixed_)),
		control(control_)
	{
		ui32_3b(length_, length);
	}
};

// WRITE FILEMARKS
class CDBWriteFilemarks : public CDB6
{
	typedef CDB6 inherited;
public:
		 uint8 opCode;
	BF4(uint8,lun:3,
				 res1:3,
				 wsmk:1,
				 immed:1);
		 uint8 count[3];
		 uint8 control;
	
	CDBWriteFilemarks(uint8 lun_, uint32 count_ = 0, bool writeSetmarks = false,
							bool immediate = false, uint8 control_ = 0)
	:	inherited(&opCode),
		opCode(SCSI_CMD_WRITE_FILEMARKS),
		B4(lun(lun_),
			res1(0),
			wsmk(writeSetmarks),
			immed(immediate)),
		control(control_)
	{
		ui32_3b(count_, count);
	}
};

//------------------------------------------------------------------------------
// Data structures returned by various requests.

// Standard INQUIRY data.
struct SCSIInquiryData : public SCSIData
{
	enum {		// Peripheral qualifiers.
		kDeviceConnected		= 0,
		kDeviceNotConnected	= 1,
		kDeviceNotSupported	= 3
	};

	enum {		// Peripheral device types.
		kDisk				= 0x00,
		kTape				= 0x01,
		kPrinter			= 0x02,
		kCPU				= 0x03,
		kWORM				= 0x04,
		kCDROM			= 0x05,
		kScanner			= 0x06,
		kOptical			= 0x07,
		kMediumChanger	= 0x08,
		kCommunication	= 0x09,
		kPrepress1		= 0x0a,
		kPrepress2		= 0x0b,
		kUnknown			= 0x1f
	};

	BF2(uint8,peripheralQualifier:3,
				 peripheralDeviceType:5);
	BF2(uint8,rmb:1,
				 deviceTypeModifier:7);
	BF3(uint8,isoVersion:2,
				 ecmaVersion:3,
				 ansiVersion:3);
	BF4(uint8,aenc:1,
				 trmiop:1,
				 res3:2,
				 responseDataFormat:4);
		 uint8 additionalLength;
		 uint8 res5;
		 uint8 res6;
	BF8(uint8,reladr:1,
				 wbus32:1,
				 wbus16:1,
				 sync:1,
				 linked:1,
				 res7:1,
				 cmdque:1,
				 sftre:1);
		 char  vendorID[8];
		 char  productID[16];
		 char  productRevisionLevel[4];
};

// REQUEST SENSE error code 0x70/0x71 sense data format.
struct SCSISenseData : public SCSIData
{
	enum {		// Error codes.
		kCurrentErrors		= 0x70,
		kDeferredErrors	= 0x71
	};
	
	enum {		// Sense keys.
		kNoSense				= 0x0,
		kRecoveredError	= 0x1,
		kNotReady			= 0x2,
		kMediumError		= 0x3,
		kHardwareError		= 0x4,
		kIllegalRequest	= 0x5,
		kUnitAttention		= 0x6,
		kDataProtect		= 0x7,
		kBlankCheck			= 0x8,
		kVendorSpecific	= 0x9,
		kCopyAborted		= 0xa,
		kAbortedCommand	= 0xb,
		kEqual				= 0xc,
		kVolumeOverflow	= 0xd,
		kMiscompare			= 0xe
	};

	BF2(uint8,valid:1,
				 errorCode:7);
		 uint8 segmentNumber;
	BF5(uint8,filemark:1,
				 eom:1,
				 ili:1,
				 res2:1,
				 senseKey:4);
		 uint8 information_[4];
		 int32 information(void) const
		 {
			 return SCSIData::s4b_i32(information_);
		 };
		 uint8 addlSenseLength;
		 uint8 cmdSpecInfo_[4];
		 uint32 cmdSpecInfo(void) const
		 {
			 return SCSIData::u4b_i32(cmdSpecInfo_);
		 };
		 uint8 asc;
		 uint8 asq;
		 uint8 fieldReplUnitCode;
	union {
		// Sense key specific field.
		uint8 senseKeySpecific[3];
		
		struct {
			BF5(uint8,sksv:1,
						 cd:1,
						 res15:2,
						 bpv:1,
						 bitPointer:3);
				 uint8 fieldPointer_[2];
				 uint16 fieldPointer(void) const
				 {
					 return SCSIData::u2b_i16(fieldPointer_);
				 }
		} illegalRequestPtr;
		
		struct {
			BF2(uint8,sksv:1,
						 res15:7);
				 uint8 retryCount_[2];
				 uint16 retryCount(void) const
				 {
					 return SCSIData::u2b_i16(retryCount_);
				 }
		} actualRetryCount;
		
		struct {
			BF2(uint8,sksv:1,
						 res15:7);
				 uint8 progressIndication_[2];
				 uint16 progressIndication(void) const
				 {
					 return SCSIData::u2b_i16(progressIndication_);
				 }
		} formatProgressIndication;
	};
};

// MODE SELECT / MODE SENSE mode parameters.
struct SCSIModeParameterHeader6 : public SCSIData
{
	enum {		// Sequential-access device buffered modes.
		kUnbuffered		= 0,
		kBuffered		= 1,
		kMultiBuffered	= 2
	};

	uint8 modeDataLength;
	uint8 mediumType;
	union {
		// Device specific parameter.
		uint8 deviceSpecific;
		
		struct {
			BF3(uint8,wp:1,
						 bufferedMode:3,
						 speed:4);
		} tape;
	};
	uint8 blockDescriptorLength;
};
	
struct SCSIBlockDescriptor : public SCSIData
{
	enum {		// Sequential-access device density codes (some of them).
		kDefaultDensity	= 0x00,
		kNoChangeDensity	= 0x7f
	};
	
	uint8	densityCode;
	uint8 numBlocks_[3];
	uint32 numBlocks(void) const
	{
		return SCSIData::u3b_i32(numBlocks_);
	}
	uint8 res4;
	uint8 blockLength_[3];
	uint32 blockLength(void) const
	{
		return SCSIData::u3b_i32(blockLength_);
	}
};

struct SCSIDeviceConfigurationPage : public SCSIData
{
	enum {		// Page code.
		kPageCode = 0x10
	};
	
	enum {		// Data compression algorithm.
		kNoCompression			= 0x00,
		kDefaultCompression	= 0x01
	};
	
	BF3(uint8,ps:1,
				 res0:1,
				 pageCode:6);
		 uint8 pageLength;
	BF6(uint8,res2:1,
				 cap:1,
				 caf:1,
				 disraw:1,
				 enc3:1,
				 ngroup:3);
		 uint8 activePartition;
		 uint8 writeBufferFullRatio;
		 uint8 readBufferEmptyRatio;
		 uint8 writeDelayTime_[2];
		 uint16 writeDelayTime(void) const
		 {
			 return SCSIData::u2b_i16(writeDelayTime_);
		 }
	BF7(uint8,dbr:1,
				 bis:1,
				 rsmk:1,
				 avc:1,
				 socf:2,
				 rbo:1,
				 rew:1);
		 uint8 gapSize;
	BF4(uint8,eodDefined:3,
				 eeg:1,
				 sew:1,
				 res10:3);
		 uint8 bufferSizeAtEW_[3];
		 uint32 bufferSizeAtEW(void) const
		 {
			 return SCSIData::u3b_i32(bufferSizeAtEW_);
		 }
		 uint8 dataCompression;
		 uint8 res15;
};

struct SCSIDataCompressionPage : public SCSIData
{
	enum {		// Page code.
		kPageCode = 0x0f
	};
	
	enum {		// Data compression algorithm.
		kNoCompression			= 0x00,
		kDefaultCompression	= 0x01,
		kDCLZCompression		= 0x20
	};
	
	BF3(uint8,ps:1,
				 res0:1,
				 pageCode:6);
		 uint8 pageLength;
	BF3(uint8,dce:1,
				 dcc:1,
				 res2:6);
	BF3(uint8,dde:1,
				 red:2,
				 res3:5);
		 uint8 res4;
		 uint8 res5;
		 uint8 res6;
		 uint8 compressionAlgorithm;
		 uint8 res8;
		 uint8 res9;
		 uint8 res10;
		 uint8 decompressionAlgorithm;
		 uint8 res12;
		 uint8 res13;
		 uint8 res14;
		 uint8 res15;
};

struct SCSIMediumPartitionPage1 : public SCSIData
{
	enum {		// Page code.
		kPageCode = 0x11
	};
	
	enum {		// Partition size unit of measure.
		kBytes	 	= 0x0,
		kKilobytes	= 0x1,
		kMegabytes	= 0x2
	};
	
	BF3(uint8,ps:1,
				 res0:1,
				 pageCode:6);
		 uint8 pageLength;
		 uint8 maximumAdditionalPartitions;
		 uint8 additionalPartitionsDefined;
	BF5(uint8,fdp:1,
				 sdp:1,
				 idp:1,
				 psum:2,
				 res4:3);
		 uint8 mediumFormatRecognition;
		 uint8 res6;
		 uint8 res7;
	struct {
		uint8 partitionSize_[2];
		uint16 partitionSize(void) const
		{
			return SCSIData::u2b_i16(partitionSize_);
		}
	} partitionSizeDescriptor[64];
};

struct SCSIMediumPartitionPage2To4 : public SCSIData
{
	BF3(uint8,ps:1,
				 res0:1,
				 pageCode:6);
		 uint8 pageLength;
	struct {
		uint8 partitionSize_[2];
		uint16 partitionSize(void) const
		{
			return SCSIData::u2b_i16(partitionSize_);
		}
	} partitionSizeDescriptor[64];
};

struct SCSIMediumPartitionPage2 : public SCSIMediumPartitionPage2To4
{
	enum {		// Page code.
		kPageCode = 0x12
	};
};
	
struct SCSIMediumPartitionPage3 : public SCSIMediumPartitionPage2To4
{
	enum {		// Page code.
		kPageCode = 0x13
	};
};
	
struct SCSIMediumPartitionPage4 : public SCSIMediumPartitionPage2To4
{
	enum {		// Page code.
		kPageCode = 0x14
	};
};
	
// READ BLOCK LIMITS data.
struct SCSIReadBlockLimitsData : public SCSIData
{
	uint8 res0;
	uint8 maxBlockLength_[3];
	uint32 maxBlockLength(void) const
	{
		return SCSIData::u3b_i32(maxBlockLength_);
	}
	uint8 minBlockLength_[2];
	uint16 minBlockLength(void) const
	{
		return SCSIData::u2b_i16(minBlockLength_);
	}
};

// READ POSITION data.
struct SCSIReadPositionData : public SCSIData
{
	BF5(uint8,bop:1,
				 eop:1,
				 res00:3,
				 bpu:1,
				 res01:2);
		 uint8 partitionNumber;
		 uint8 res2;
		 uint8 res3;
	uint8 firstBlockLocation_[4];
	uint32 firstBlockLocation(void) const
	{
		return SCSIData::u4b_i32(firstBlockLocation_);
	}
	uint8 lastBlockLocation_[4];
	uint32 lastBlockLocation(void) const
	{
		return SCSIData::u4b_i32(lastBlockLocation_);
	}
	uint8 res12;
	uint8 numBlocksInBuffer_[3];
	uint32 numBlocksInBuffer(void) const
	{
		return SCSIData::u3b_i32(numBlocksInBuffer_);
	}
	uint8 numBytesInBuffer_[4];
	uint32 numBytesInBuffer(void) const
	{
		return SCSIData::u4b_i32(numBytesInBuffer_);
	}
};

#endif // ST_SCSIDATA_H
